@@ -2,6 +2,7 @@
 #define _GEOS_H
 
 #include <Python.h>
+#include <numpy/ndarraytypes.h>
 
 /* To avoid accidental use of non reentrant GEOS API. */
 #ifndef GEOS_USE_ONLY_R_API
@@ -52,8 +53,11 @@ enum {
   PGERR_GEOMETRY_TYPE,
   PGERR_MULTIPOINT_WITH_POINT_EMPTY,
   PGERR_EMPTY_GEOMETRY,
+  PGERR_GEOJSON_EMPTY_POINT,
+  PGERR_LINEARRING_NCOORDS,
   PGWARN_INVALID_WKB,  // raise the GEOS WKB error as a warning instead of exception
-  PGWARN_INVALID_WKT   // raise the GEOS WKB error as a warning instead of exception
+  PGWARN_INVALID_WKT,  // raise the GEOS WKT error as a warning instead of exception
+  PGWARN_INVALID_GEOJSON
 };
 
 // Define how the states are handled by CPython
@@ -87,6 +91,14 @@ enum {
     case PGERR_EMPTY_GEOMETRY:                                                           \
       PyErr_SetString(PyExc_ValueError, "One of the Geometry inputs is empty.");         \
       break;                                                                             \
+    case PGERR_GEOJSON_EMPTY_POINT:                                                      \
+      PyErr_SetString(PyExc_ValueError,                                                  \
+                      "GeoJSON output of empty points is currently unsupported.");       \
+      break;                                                                             \
+    case PGERR_LINEARRING_NCOORDS:                                                       \
+      PyErr_SetString(PyExc_ValueError,                                                  \
+                      "A linearring requires at least 4 coordinates.");                  \
+      break;                                                                             \
     case PGWARN_INVALID_WKB:                                                             \
       PyErr_WarnFormat(PyExc_Warning, 0,                                                 \
                        "Invalid WKB: geometry is returned as None. %s", last_error);     \
@@ -95,6 +107,10 @@ enum {
       PyErr_WarnFormat(PyExc_Warning, 0,                                                 \
                        "Invalid WKT: geometry is returned as None. %s", last_error);     \
       break;                                                                             \
+    case PGWARN_INVALID_GEOJSON:                                                         \
+      PyErr_WarnFormat(PyExc_Warning, 0,                                                 \
+                       "Invalid GeoJSON: geometry is returned as None. %s", last_error); \
+      break;                                                                             \
     default:                                                                             \
       PyErr_Format(PyExc_RuntimeError,                                                   \
                    "Pygeos ufunc returned with unknown error state code %d.", errstate); \
@@ -141,12 +157,14 @@ extern PyObject* geos_exception[1];
 extern void geos_error_handler(const char* message, void* userdata);
 extern void geos_notice_handler(const char* message, void* userdata);
 extern void destroy_geom_arr(void* context, GEOSGeometry** array, int length);
-#if !GEOS_SINCE_3_10_0
 extern char has_point_empty(GEOSContextHandle_t ctx, GEOSGeometry* geom);
 extern GEOSGeometry* point_empty_to_nan_all_geoms(GEOSContextHandle_t ctx,
                                                   GEOSGeometry* geom);
-#endif  // !GEOS_SINCE_3_10_0
 extern char check_to_wkt_compatible(GEOSContextHandle_t ctx, GEOSGeometry* geom);
+#if GEOS_SINCE_3_9_0
+extern char wkt_empty_3d_geometry(GEOSContextHandle_t ctx, GEOSGeometry* geom,
+                                  char** wkt);
+#endif  // GEOS_SINCE_3_9_0
 extern char geos_interpolate_checker(GEOSContextHandle_t ctx, GEOSGeometry* geom);
 
 extern int init_geos(PyObject* m);
@@ -156,5 +174,11 @@ int get_bounds(GEOSContextHandle_t ctx, GEOSGeometry* geom, double* xmin, double
 GEOSGeometry* create_box(GEOSContextHandle_t ctx, double xmin, double ymin, double xmax,
                          double ymax, char ccw);
 GEOSGeometry* create_point(GEOSContextHandle_t ctx, double x, double y);
+GEOSGeometry* PyGEOSForce2D(GEOSContextHandle_t ctx, GEOSGeometry* geom);
+GEOSGeometry* PyGEOSForce3D(GEOSContextHandle_t ctx, GEOSGeometry* geom, double z);
+
+GEOSCoordSequence* coordseq_from_buffer(GEOSContextHandle_t ctx, const double* buf,
+                                        unsigned int size, unsigned int dims, char ring_closure,
+                                        npy_intp cs1, npy_intp cs2);
 
 #endif  // _GEOS_H
